import argparse
import os
from typing import List

from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter
from wai.logging import LOGGING_WARNING
from wai.spectralio.asc import Writer as SWriter

from sdc.api import Spectrum2D, SplittableStreamWriter, make_list


class ASCWriter(SplittableStreamWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_dir: str = None, instrument_name: str = None, accessory_name: str = None,
                 data_points: int = None, first_x_point: float = None, last_x_point: float = None, descending: bool = None,
                 split_names: List[str] = None, split_ratios: List[int] = None, split_group: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param output_dir: the output directory to save the .spec files in
        :type output_dir: str
        :param instrument_name: the instrument name used in the header
        :type instrument_name: str
        :param accessory_name: the accessory name used in the header
        :type accessory_name: str
        :param data_points: the number of data points to use, -1 for all
        :type data_points: int
        :param first_x_point: the first wave number
        :type first_x_point: float
        :param last_x_point: the last wave number
        :type last_x_point: float
        :param descending: whether to output the wave numbers in descending order
        :type descending: bool
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param split_group: the regular expression with a single group used for keeping items in the same split, e.g., for identifying the base name of a file or the sample ID
        :type split_group: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, split_group=split_group, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.instrument_name = instrument_name
        self.accessory_name = accessory_name
        self.data_points = data_points
        self.first_x_point = first_x_point
        self.last_x_point = last_x_point
        self.descending = descending

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-asc"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the spectrum in ASC format."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the .asc files in. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        parser.add_argument("--instrument_name", type=str, help="The instrument name to use in the header", required=False, default="<not implemented>")
        parser.add_argument("--accessory_name", type=str, help="The accessory name to use in the header", required=False, default="ABB-BOMEM MB160D")
        parser.add_argument("--data_points", type=int, help="The number of data points to output, -1 for all", required=False, default=-1)
        parser.add_argument("--first_x_point", type=float, help="The first wave number", required=False, default=3749.3428948242)
        parser.add_argument("--last_x_point", type=float, help="The last wave number", required=False, default=9998.2477195313)
        parser.add_argument("--descending", action="store_true", help="Outputs the wave numbers in descending order")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.instrument_name = ns.instrument_name
        self.accessory_name = ns.accessory_name
        self.data_points = ns.data_points
        self.first_x_point = ns.first_x_point
        self.last_x_point = ns.last_x_point
        self.descending = ns.descending

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [Spectrum2D]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.instrument_name is None:
            self.instrument_name = "<not implemented>"
        if self.accessory_name is None:
            self.accessory_name = "ABB-BOMEM MB160D"
        if self.data_points is None:
            self.data_points = -1
        if self.first_x_point is None:
            self.first_x_point = 3749.3428948242
        if self.last_x_point is None:
            self.last_x_point = 9998.2477195313
        if self.descending is None:
            self.descending = False

    def _compile_options(self) -> List[str]:
        """
        Compiles the options for initializing the underlying writer.

        :return: the list of options to use
        :rtype: list
        """
        result = super()._compile_options()
        result.extend(["--instrument-name", self.instrument_name])
        result.extend(["--accessory-name", self.accessory_name])
        result.append("--data-points=" + str(self.data_points))
        result.extend(["--first-x-point", str(self.first_x_point)])
        result.extend(["--last-x-point", str(self.last_x_point)])
        if self.descending:
            result.append("--descending")
        return result

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        writer = SWriter()
        writer.options = self._compile_options()

        for item in make_list(data):
            sub_dir = self.session.expand_placeholders(self.output_dir)
            if self.splitter is not None:
                split = self.splitter.next(item=item.spectrum_name)
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating dir: %s" % sub_dir)
                os.makedirs(sub_dir)

            path = os.path.join(sub_dir, item.spectrum_name)
            path = os.path.splitext(path)[0] + ".asc"
            self.logger().info("Writing spectrum to: %s" % path)
            writer.write([item.spectrum], path)
